"""Module for defining Macro classes."""

import shlex


class MacroFactory:
    """Generates Macro objects based on the provided commands.

    :param list[str] commands: The commands to use for building Macros.
    :param list[str] prefixes: The command prefixes to use for building Macros.
    :param list[str] suffixes: The command suffixes to use for building Macros.
    """

    __slots__ = ['_commands']

    def __init__(self, commands, prefixes=None, suffixes=None):
        """Instantiates a new MacroFactory object."""
        if not prefixes:
            prefixes = []
        if not suffixes:
            suffixes = []
        if commands is None:
            self._commands = []
        else:
            while len(prefixes) < len(commands):
                prefixes.append('')
            while len(suffixes) < len(commands):
                suffixes.append('')
            self._commands = tuple(zip(prefixes, commands, suffixes))

    def generate(self):
        """Generates Macro objects based on the commands supplied to the MacroFactory.

        :rtype: list[Macro]
        :return: A list of generated Macro objects.
        """
        macros = []
        i = 0
        for prefix, cmd, suffix, in self._commands:
            if not cmd:
                continue
            macros.append(
                Macro(cmd, sequence=i, prefix=prefix, suffix=suffix)
            )
            i += 1
        return macros


class Macro:
    """A command to be executed by a CommandRunner.

    Macros are the primary object for doing work in build-magic. A Macro is composed of a command
    to be executed at the command-line by a CommandRunner.

    :param str command: The base command to be executed as supplied by the user.
    :param int sequence: The run order of the command.
    :param str prefix: A command or portion of a command to append to the beginning of the base command.
    :param str suffix: A command or portion of a command to append to the end of the base command.
    """

    __slots__ = ['_command', 'prefix', 'sequence', 'suffix']

    def __init__(self, command='', sequence=0, prefix='', suffix=''):
        """Instantiates a new Macro object."""
        if not isinstance(command, str):
            raise TypeError('command must by str not {}'.format(type(command)))
        self._command = command
        self.sequence = int(sequence)
        self.prefix = str(prefix)
        self.suffix = str(suffix)

    @property
    def command(self):
        """Provides the command to execute.

        :rtype: str
        :return: The Macro command.
        """
        return self._command

    def as_list(self):
        """Breaks up a command into a list.

        :rtype: list[str]
        :return: The command as a list.
        """
        def prep(cmd):
            if cmd:
                cmd = shlex.split(cmd)
            else:
                cmd = []
            return cmd

        command = prep(self.prefix) + prep(self._command) + prep(self.suffix)
        return command

    def as_string(self):
        """Provides a command as a string.

        :rtype: str
        :return: The command as a string.
        """
        cmd = self._command
        if self.prefix:
            cmd = self.prefix + ' ' + cmd
        if self.suffix:
            cmd += ' ' + self.suffix
        return cmd



"""Module for defining build-magic output functionality."""

from datetime import datetime
import os
import sys

from colorama import Cursor, Fore, init, Style

from build_magic import __version__ as version
from build_magic.reference import OutputMethod


class Output:
    """Interface for defining output methods."""

    def __init__(self):
        """Instantiates a new Output class."""
        self.timer = None

    def start_job(self, *args, **kwargs):
        """Indicates the beginning of a sequence of stages."""
        raise NotImplementedError

    def end_job(self, *args, **kwargs):
        """Indicates the end of a build-magic session."""
        raise NotImplementedError

    def start_stage(self, *args, **kwargs):
        """Indicates the beginning of a stage."""
        raise NotImplementedError

    def end_stage(self, *args, **kwargs):
        """Indicates the end of a stage."""
        raise NotImplementedError

    def no_job(self, *args, **kwargs):
        """Indicates there are no commands to execute."""
        raise NotImplementedError

    def macro_start(self, *args, **kwargs):
        """Indicates the beginning of a command."""
        raise NotImplementedError

    def macro_status(self, *args, **kwargs):
        """Indicates the success status of a command."""
        raise NotImplementedError

    def error(self, *args, **kwargs):
        """Communicates an error message."""
        raise NotImplementedError

    def info(self, *args, **kwargs):
        """Communicates a general information message."""
        raise NotImplementedError

    @staticmethod
    def _display(line, err=False):
        """Prints a message to stdout.

        :param str line: The message to print.
        :param bool err: If True, output will be written to sys.stderr.
        :return: None
        """
        if err:
            print(line, file=sys.stderr)
        else:
            print(line)

    def log(self, method, *args, **kwargs):
        """Generic method for calling valid output methods.

        :param str|build_magic.reference.OutputMethod method: The output method to use.
        :param args: Optional arguments to pass to the output method.
        :param kwargs: Optional kwargs to pass to the output method.
        :return: None
        """
        if isinstance(method, OutputMethod):
            method = method.value
        func = getattr(self, method)
        func(*args, **kwargs)

    def print_output(self, message, is_error=False):
        """High level method for printing a message to stdout.

        :param str|bytes message: The message to print to stdout.
        :param bool is_error: Prints using the error format if True, otherwise prints using the info format.
        :return: None
        """
        level = OutputMethod.ERROR if is_error else OutputMethod.INFO
        if isinstance(message, bytes):
            out = message.decode('utf-8')
        else:
            out = str(message)
        self.log(level, out)


class Basic(Output):
    """Prototype output to the commandline used for testing."""

    def start_job(self):
        """Indicates the beginning of a sequence of stages."""
        message = f'{datetime.now().isoformat()} build-magic [ INFO  ] version {version}'
        self._display(message)
        self.timer = datetime.now()

    def end_job(self):
        """Indicates the end of a build-magic session.

        :return: None
        """
        if self.timer:
            delta = datetime.now() - self.timer
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] finished in {delta.total_seconds():.3f}'
        else:
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] finished'
        self._display(message)

    def start_stage(self, stage_number=1, name=None):
        """Indicates the beginning of a stage.

        :param int stage_number: The stage sequence number.
        :param str|None name: The stage name if given.
        :return: None
        """
        if name:
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] Starting Stage {stage_number}: {name}'
        else:
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] Starting Stage {stage_number}'
        self._display(message)

    def end_stage(self, stage_number=1, status_code=0, name=None):
        """Indicates the end of a stage.

        :param int stage_number: The stage sequence number.
        :param int status_code: The highest exit code for the stage.
        :param str|None name: The stage name if given.
        :return: None
        """
        result = 'DONE'
        if status_code > 0:
            result = 'FAIL'
        if name:
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] Stage {stage_number}: {name} - ' \
                      f'complete with result {result}'
        else:
            message = f'{datetime.now().isoformat()} build-magic [ INFO  ] Stage {stage_number} ' \
                      f'complete with result {result}'
        self._display(message)

    def no_job(self):
        """Indicates there are no commands to execute."""
        self._display('No commands to run. Use --help for usage. Exiting...')

    def macro_start(self, *args, **kwargs):
        """Not used by the Basic Output class."""
        return

    def macro_status(self, directive, command='', status_code=0):
        """Indicates the success status of a command.

        :param str directive: The executed macro's directive.
        :param str command: The executed command.
        :param status_code: The macro's exit code.
        :return: None
        """
        result = 'DONE'
        if status_code > 0:
            result = 'FAIL'
        if not command:
            message = f'{datetime.now().isoformat()} build-magic [ {result:<6}] {directive.upper():<8}'
        else:
            message = f'{datetime.now().isoformat()} build-magic [ {result:<6}] {directive.upper():<8} : {command}'
        self._display(message)

    def error(self, err):
        """Communicates an error message.

        :param str err: The error message to display.
        :return: None
        """
        message = f'{datetime.now().isoformat()} build-magic [ ERROR ] {err}'
        self._display(message)

    def info(self, msg):
        """Communicates a general information message.

        :param str msg: The message to print.
        :return: None
        """
        message = f'{datetime.now().isoformat()} build-magic [ INFO  ] OUTPUT   : {msg}'
        self._display(message)


class Tty(Output):
    """Fancy output when using a Terminal."""

    def __init__(self):
        """Instantiates a new Tty class."""
        super().__init__()
        init()

    @staticmethod
    def get_width():
        """If using a terminal, get the width, otherwise set it manually."""
        try:
            term = os.get_terminal_size()
            return term.columns
        except OSError:
            return 80

    @staticmethod
    def get_height():
        """If using a terminal, get the height, otherwise set it manually."""
        try:
            term = os.get_terminal_size()
            return term.lines
        except OSError:
            return 20

    def start_job(self):
        """Indicates the beginning of a sequence of stages."""
        # Create the hammer and sparkle glyphs only if using a TTY and on Mac or Linux.
        emoji = '\U0001f528' + '\U00002727' if sys.stdout.isatty() and sys.platform != 'win32' else ''
        message = Fore.CYAN + Style.BRIGHT + 'build-magic{} {}\n'.format(emoji, version) + Style.RESET_ALL
        message += Fore.CYAN + 'Start time {}\n'.format(datetime.now().strftime('%c')) + Style.RESET_ALL
        self._display(message)
        self.timer = datetime.now()

    def end_job(self):
        """Indicates the end of a build-magic session.

        :return: None
        """
        if self.timer:
            delta = datetime.now() - self.timer
            message = 'build-magic finished in {:.3f} seconds'.format(delta.total_seconds())
        else:
            message = 'build-magic finished at {}'.format(datetime.now().strftime('%c'))
        self._display(message)

    def start_stage(self, stage_number=1, name=None):
        """Indicates the beginning of a stage.

        :param int stage_number: The stage sequence number.
        :param str|None name: The stage name if given.
        :return: None
        """
        if name:
            message = f'Starting Stage {stage_number}: {name}'
        else:
            message = f'Starting Stage {stage_number}'
        self._display(message)

    def end_stage(self, stage_name=1, status_code=0, name=None):
        """Indicates the end of a stage.

        :param int stage_name: The stage sequence number.
        :param int status_code: The highest exit code for the stage.
        :param str|None name: The stage name if given.
        :return: None
        """
        color = Fore.GREEN + Style.BRIGHT
        result = 'COMPLETE'
        if status_code > 0:
            color = Fore.RED + Style.BRIGHT
            result = 'FAILED'
        if name:
            message = f'Stage {stage_name}: {name} - finished with result {color}{result}{Style.RESET_ALL}\n'
        else:
            message = f'Stage {stage_name} finished with result {color}{result}{Style.RESET_ALL}\n'
        self._display(message)

    def no_job(self):
        """Indicates there are no commands to execute."""
        self._display(Fore.YELLOW + 'No commands to run. Use --help for usage. Exiting...')

    def macro_start(self, directive, command=''):
        """Indicates the start of a command.

        :param str directive: The directive of the executing command.
        :param str command: The current executing command.
        :return: None
        """
        width = self.get_width()
        status = Fore.YELLOW + Style.BRIGHT + 'RUNNING' + Style.RESET_ALL
        spacing = width - 22 - len(command)
        if len(command) + 11 > width - 11:
            command = command[:width - 23] + '....'
        if not command:
            message = '{} {} {}'.format(directive.upper(), command, '.' * spacing, status)
        else:
            message = '{:<8}: {} {} {}'.format(directive.upper(), command, '.' * spacing, status)
        self._display(message)

    def macro_status(self, directive='', command='', status_code=0):
        """Indicates the success status of a command.

        :param str directive: The executed macro's directive.
        :param str command: The executed command.
        :param status_code: The macro's exit code.
        :return: None
        """
        width = self.get_width()
        position = width - 10
        height = self.get_height() - 1
        if status_code > 0:
            result = Fore.RED + Style.BRIGHT + '{:<8}'.format('FAILED') + Style.RESET_ALL
        else:
            result = Fore.GREEN + Style.BRIGHT + '{:<8}'.format('COMPLETE') + Style.RESET_ALL
        self._display(Cursor.POS(position, height) + result)

    def error(self, err):
        """Communicates an error message.

        :param str err: The error message to display.
        :return: None
        """
        width = self.get_width()
        position = width - 10
        height = self.get_height() - 1
        result = Fore.RED + Style.BRIGHT + '{:<8}'.format('ERROR') + Style.RESET_ALL
        self._display(Cursor.POS(position, height) + result)
        self._display(Fore.RED + str(err) + Style.RESET_ALL, err=True)

    def info(self, msg):
        """Communicates a general information message.

        :param str msg: The message to print.
        :return: None
        """
        message = 'OUTPUT  : {}'.format(msg)
        self._display(message)


class Silent(Output):
    """Silent output that suppresses output."""

    def start_job(self, *args, **kwargs):
        """Indicates the beginning of a sequence of stages."""
        return

    def end_job(self, *args, **kwargs):
        """Indicates the end of a build-magic session."""
        return

    def start_stage(self, *args, **kwargs):
        """Indicates the beginning of a stage."""
        return

    def end_stage(self, *args, **kwargs):
        """Indicates the end of a stage."""
        return

    def no_job(self, *args, **kwargs):
        """Indicates there are no commands to execute."""
        return

    def macro_start(self, *args, **kwargs):
        """Indicates the beginning of a command."""
        return

    def macro_status(self, *args, **kwargs):
        """Indicates the success status of a command."""
        return

    def error(self, *args, **kwargs):
        """Communicates an error message."""
        return

    def info(self, *args, **kwargs):
        """Communicates a general information message."""
        return



"""Click CLI for running build-magic."""

import logging
import sys

import click
import yaml

# Need this to disable logging in the vagrant package when Vagrant isn't installed.
logging.disable(logging.WARNING)

from build_magic import __version__ as ver
from build_magic import core
from build_magic.exc import ExecutionError, NoJobs, SetupError, TeardownError
from build_magic import reference

# Get a list of command runners.
RUNNERS = click.Choice(reference.Runners.available(), case_sensitive=False)


# Get a list of available actions.
ACTIONS = click.Choice(reference.Actions.available(), case_sensitive=False)


# Defines the type for the working directory parameter.
WORKINGDIR = click.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=False, allow_dash=False)


# Defines the type for the config file.
CONFIG = click.File(mode='r', encoding='utf-8', errors='strict', lazy=False)


USAGE = """Usage: build-magic [OPTIONS] [ARGS]...

build-magic is an un-opinionated build automation tool. Some potential uses include:
    * Building applications across multiple platforms.
    * Conducting installation dry runs.
    * Testing application installs across multiple platforms.
    * Deploying and installing artifacts to remote machines.

Examples:
* Archive two files on the local machine.
    build-magic tar -czf myfiles.tar.gz file1.txt file2.txt

* Archive two files on the local machine and delete the original files.
    build-magic -c build "tar -czf myfiles.tar.gz file1.txt file2.txt" -c execute "rm file1.txt file2.txt"
    
* Copy two files to a remote machine and archive them.
    build-magic -r remote -e user@myhost --copy . -c build "tar -czf myfiles.tar.gz f1.txt f2.txt" f1.txt f2.txt

* Build a project in a Linux container.
    build-magic -r docker -e Ubuntu:latest -c build "make all"

Use --help for detailed usage of each option.

Visit https://cmmorrow.github.io/build-magic/user_guide/cli_usage/ for a detailed usage description.
"""


@click.command()
@click.option('--command', '-c', help='A directive, command pair to execute.', multiple=True, type=(str, str))
@click.option('--config', '-C', help='The config file to load parameters from.', multiple=True, type=CONFIG)
@click.option('--copy', help='Copy from the specified path.', default='', type=str)
@click.option('--environment', '-e', help='The command runner environment to use.', default='', type=str)
@click.option('--runner', '-r', help='The command runner to use.', type=RUNNERS)
@click.option('--name', help='The stage name to use.', type=str)
@click.option('--wd', help='The working directory to run commands from.', default='.', type=WORKINGDIR)
@click.option('--continue/--stop', 'continue_', help='Continue to run after failure if True.', default=False)
@click.option('--parameter', '-p', help='Key/value used for runner specific settings.', multiple=True, type=(str, str))
@click.option('--action', help='Setup and teardown action to perform.', type=ACTIONS)
@click.option('--plain/--fancy', help='Enable basic output. Ideal for automation.', default=False)
@click.option('--quiet', help='Suppress all output from build-magic.', is_flag=True)
@click.option('--verbose', help='Verbose output -- stdout from executed commands will be printed.', is_flag=True)
@click.option('--version', help='Show the version and exit.', is_flag=True)
@click.argument('args', nargs=-1)
def build_magic(
        command,
        config,
        copy,
        continue_,
        environment,
        args,
        parameter,
        action,
        runner,
        name,
        wd,
        plain,
        quiet,
        verbose,
        version,
):
    """An un-opinionated build automation tool.

    ARGS - Files as arguments to copy from the copy path to the working directory.
    Alternatively, ARGS can be a single command to execute if the --command option isn't used.

    Visit https://cmmorrow.github.io/build-magic/user_guide/cli_usage/ for a detailed usage description.
    """
    if version:
        click.echo(ver)
        sys.exit(0)

    # Get the output type.
    if plain:
        out = reference.OutputTypes.BASIC
    elif quiet:
        out = reference.OutputTypes.SILENT
    else:
        out = reference.OutputTypes.TTY

    stages_ = []

    if config:
        seq = core.iterate_sequence()
        for cfg in config:
            # Read the config YAML file.
            obj = yaml.safe_load(cfg)

            # Parse the YAML file and set the options.
            try:
                stages = core.config_parser(obj)
            except ValueError as err:
                click.secho(str(err), fg='red', err=True)
                sys.exit(reference.ExitCode.INPUT_ERROR)

            for stage_ in stages:
                stages_.append(
                    dict(
                        sequence=next(seq),
                        runner_type=stage_['runner_type'],
                        directives=stage_['directives'],
                        artifacts=stage_['artifacts'],
                        action=stage_['action'],
                        commands=stage_['commands'],
                        environment=stage_['environment'],
                        copy=stage_['copy'],
                        wd=stage_['wd'],
                        name=stage_['name'],
                        parameters=stage_['parameters'],
                    )
                )
    else:
        # Set the commands from the command line.
        if command:
            directives, commands = list(zip(*command))
            artifacts = args
        else:
            directives, commands = ['execute'], [' '.join(args)]
            artifacts = []
        if not commands or commands == ['']:
            click.echo(USAGE)
            sys.exit(5)

        stages_.append(
            dict(
                sequence=1,
                runner_type=reference.Runners.LOCAL.value,
                directives=directives,
                artifacts=artifacts,
                action=reference.Actions.DEFAULT.value,
                commands=commands,
                environment=environment,
                copy=copy,
                wd=wd,
                parameters=parameter,
            )
        )
        if name:
            stages_[0].update(dict(name=name))

    # Override values in the config file with options set at the command line.
    for stage in stages_:
        if action:
            stage.update(dict(action=action))
        if environment:
            stage.update(dict(environment=environment))
        if copy:
            stage.update(dict(copy=copy))
        if len(wd) > 1:
            stage.update(dict(wd=wd))
        if runner:
            stage.update(dict(runner_type=runner))

    # Build the stages.
    stages = []
    for stage in stages_:
        try:
            stages.append(core.build_stage(**stage))
        except (NotADirectoryError, ValueError, reference.ValidationError) as err:
            click.secho(str(err), fg='red', err=True)
            sys.exit(reference.ExitCode.INPUT_ERROR)

    # Run the stages.
    try:
        engine = core.Engine(stages, continue_on_fail=continue_, output_format=out, verbose=verbose)
        code = engine.run()
    except NoJobs:
        sys.exit(reference.ExitCode.NO_TESTS)
    except (ExecutionError, SetupError, TeardownError):
        sys.exit(reference.ExitCode.INTERNAL_ERROR)
    except KeyboardInterrupt:
        click.secho('\nbuild-magic interrupted and exiting....', fg='red', err=True)
        sys.exit(reference.ExitCode.INTERRUPTED)

    sys.exit(code)



"""Module for defining build-magic core classes and exceptions."""


from pathlib import Path
from pkg_resources import resource_filename
import types

import json
from jsonschema import ValidationError, validate as jsvalidator

from build_magic import actions, output, runner
from build_magic.exc import ExecutionError, SetupError, TeardownError, NoJobs
from build_magic.macro import MacroFactory
from build_magic.reference import Actions, Directive, ExitCode, OutputMethod, OutputTypes, Runners
from build_magic.reference import BindDirectory, HostWorkingDirectory, KeyPassword, KeyPath, KeyType

# Add valid Parameter classes here.
PARAMETERS = (
    BindDirectory,
    HostWorkingDirectory,
    KeyPath,
    KeyType,
    KeyPassword,
)


mode = OutputMethod
_output = output.Tty()


def build_stage(*args, **kwargs):
    """Helper function for building stage objects."""
    return StageFactory.build(*args, **kwargs)


def iterate_sequence():
    """Increments the output sequence by one each time it's called.

    :rtype: Iterator[int]
    :return: The iterative sequence value.
    """
    seq = 1
    while True:
        yield seq
        seq += 1


def config_parser(config):
    """Parse the parameters from a build-magic config file.

    :param dict config: The content of the config file to parse.
    :rtype: list[dict]
    :return: A list of stage parameters.
    """
    # Read the config schema.
    schema = Path(resource_filename('build_magic', 'core.py')).parent / 'static' / 'config_schema.json'
    with open(schema, 'r') as file:
        schema = json.load(file)

    # Validate the config file.
    try:
        jsvalidator(config, schema=schema)
    except ValidationError as err:
        raise ValueError('Config validation failed: {}'.format(err))

    # Build the stages.
    stages = []
    for data in config.get('build-magic', []):
        data = data.get('stage', {})
        stage = dict()
        stage['name'] = data.get('name', '')
        stage['runner_type'] = data.get('runner', Runners.LOCAL.value)
        stage['environment'] = data.get('environment', '')
        stage['continue'] = data.get('continue on fail', False)
        stage['wd'] = data.get('working directory', '.')
        stage['copy'] = data.get('copy from directory', '')
        stage['artifacts'] = data.get('artifacts', [])

        # Set the action.
        stage['action'] = data.get('action', Actions.DEFAULT.value)

        # Set the command and directives.
        macros = data.get('commands', [])
        commands = []
        directives = []
        for macro in macros:
            directives.append(list(macro.keys())[0])
            commands.append(list(macro.values())[0])
        stage['commands'] = commands
        stage['directives'] = directives

        # Set the parameters.
        parameters = data.get('parameters', [])
        stage['parameters'] = [tuple(param.items())[0] for param in parameters]
        stages.append(stage)

    return stages


class Engine:
    """The primary driver in build-magic. The engine executes stages and reports on the results.

    :param bool continue_on_fail: If True, all jobs will attempt to run, even if the previous job failed.
        Otherwise, execution will end after a failed job.
    :param list[Stage]|None stages: The stage or stages to execute.
    """

    __slots__ = ['_continue_on_fail', '_stages', '_verbose']

    def __init__(self, stages=None, continue_on_fail=False, output_format=OutputTypes.TTY, verbose=False):
        """Executes stages and reports the results.

        :param list[Stage]|None stages: The stage or stages to execute.
        :param bool continue_on_fail: If True, continue command execution even if a command fails or errors.
        :param OutputTypes output_format: The output interface to use for displaying messages.
        :param bool verbose: If True, print the stdout from each Macro status.
        :return: The highest status code reported by a stage.
        """
        self._continue_on_fail = continue_on_fail
        self._verbose = verbose
        self._stages = stages or []

        # TODO: Check to make sure stages is a list.

        # Sort stages by sequence.
        if len(stages) > 1:
            self._stages = sorted(stages, key=lambda s: s.sequence)

        global _output
        _output = getattr(output, output_format.value)()

    def run(self):
        """Executes stages and reports the results.

        :return: The highest status code reported by a stage.
        """
        # Initialize the status code to 0.
        status_code = ExitCode.PASSED.value

        # Start
        _output.log(mode.JOB_START)

        for stage in self._stages:
            # Run the stage.
            _output.log(mode.STAGE_START, stage.sequence, stage.name)
            stage.setup()
            try:
                exit_code = stage.run(self._continue_on_fail, self._verbose)
            except (SetupError, ExecutionError, TeardownError) as err:
                exit_code = ExitCode.INTERNAL_ERROR
                _output.log(mode.ERROR, err)
                _output.log(mode.STAGE_END, stage.sequence, exit_code)
                _output.log(mode.JOB_END)
                raise err

            if exit_code > status_code:
                status_code = exit_code
            _output.log(mode.STAGE_END, stage.sequence, exit_code, stage.name)

        # TODO: This is the wrong status code - fix it.
        _output.log(mode.JOB_END)
        return status_code


class StageFactory:
    """Validates and generates Stage objects."""

    @classmethod
    def _build_macros(cls, commands):
        """Build Macro objects from provided commands.

        :param list[str] commands: The commands to use for building the Macro objects.
        :rtype: list[Macro]
        :return: A list of Macro objects.
        """
        factory = MacroFactory(commands)
        return factory.generate()

    @classmethod
    def _build_command_runner(cls, runner_type, environment, parameters, copy, wd, artifacts):
        """Build a CommandRunner object.

        :param str runner_type: The command runner to use for executing commands.
        :param str environment: The Stage environment to use.
        :param str copy: The directory to copy artifacts from, to the working directory.
        :param str wd: The working directory to use for executing commands.
        :param list[str] artifacts: The Stage artifacts to work on.
        :param dict parameters: The user-supplied parameters to pass to the command runner.
        :rtype: CommandRunner
        :return: The CommandRunner object to use for executing the Stage's commands.
        """
        if runner_type not in Runners.available():
            raise ValueError('Runner must be one of {}.'.format(', '.join(Runners.available())))

        if runner_type == Runners.VAGRANT.value and not environment:
            raise ValueError('Environment must be a path to a Vagrant file if using the Vagrant runner.')

        if runner_type == Runners.DOCKER.value and not environment:
            raise ValueError('Environment must be a Docker image if using the Docker runner.')

        if copy and not Path(copy).exists():
            raise NotADirectoryError(f'Path {copy} does not exist.')

        command_runner = getattr(runner, runner_type.capitalize())
        return command_runner(environment, working_dir=wd, copy_dir=copy, artifacts=artifacts, parameters=parameters)

    @classmethod
    def _build_parameters(cls, parameters):
        """Build the Parameter objects for the CommandRunner to use.

        :param list[tuple] parameters: The parameter keys and their corresponding classes.
        :rtype: dict[str, build_magic.reference.Parameter]
        :return: A dictionary of Parameter objects instantiated with the passed in values.
        """
        if not parameters:
            return {}
        parameter_map = dict([(p.ALIAS, p) if p.ALIAS else (p.KEY, p) for p in PARAMETERS])
        params = {}
        for param_key, param_value in parameters:
            if param_key not in parameter_map.keys():
                raise ValueError(f'Parameter {param_key} is not a valid parameter.')
            param = parameter_map[param_key]
            # Instantiating param can fail with a reference.ValidationError.
            params[param_key] = param(param_value)
        return params

    @classmethod
    def build(
            cls,
            sequence,
            runner_type,
            directives,
            artifacts,
            commands,
            environment,
            action,
            copy,
            wd,
            name=None,
            parameters=None,
    ):
        """Validates inputs and generates a new Stage object.

        :param int sequence: The sequence order to run the stage in.
        :param str runner_type: The command runner to use for executing commands.
        :param list[str] directives: The Stage directives to execute.
        :param list[str] artifacts: The Stage artifacts to work on.
        :param list[str] commands: The Stage commands to execute.
        :param str environment: The Stage environment to use.
        :param str action: The Stage action to use for execution.
        :param str copy: The directory to copy artifacts from, to the working directory.
        :param str wd: The working directory to use for executing commands.
        :param str|None name: The stage name if provided.
        :param list[tuple]|None parameters: The optional parameters passed by the user.
        :rtype: Stage
        :return: The generated Stage object.
        """
        if not commands:
            _output.log(mode.NO_JOB)
            raise NoJobs

        for directive in directives:
            directive = directive.lower()
            if directive not in Directive.available():
                raise ValueError(
                    'Directive must be one of {}'.format(', '.join(Directive.available()))
                )

        if len(commands) != (len(directives)):
            raise ValueError('Length of commands unequal to length of directives.')

        if action not in Actions.available():
            raise ValueError('Action must be one of {}.'.format(', '.join(Actions.available())))

        # Build the parameters.
        params = cls._build_parameters(parameters)

        if not artifacts:
            artifacts = []

        # Build the macros.
        macros = cls._build_macros(commands=commands)
        if not macros:
            raise ValueError('There are no commands to execute.')

        # Create the CommandRunner.
        cmd_runner = cls._build_command_runner(
            runner_type,
            environment,
            copy=copy,
            wd=wd,
            artifacts=artifacts,
            parameters=params,
        )

        return Stage(cmd_runner, macros, directives, sequence, action, name)


class Stage:
    """A bundle of commands to execute in order, with a particular CommandRunner, environment, and action."""

    __slots__ = [
        '_action',
        '_command_runner',
        '_directives',
        '_is_setup',
        '_macros',
        '_result',
        '_results',
        '_sequence',
        '_name',
    ]

    def __init__(self, cmd_runner, macros, directives, sequence, action, name=None):
        """Instantiates a new Stage object.

        Note: Stage objects should not be constructed directly and should instead be created by a StageFactory.

        :param CommandRunner cmd_runner: The CommandRunner to use.
        :param list[Macro] macros: The commands to execute.
        :param list[str] directives: The command directives.
        :param int|str sequence: The execution order of the macros.
        :param str action: The Action to use.
        :param str|None name: The stage name if provided.
        """
        try:
            self._action = getattr(actions, action.capitalize())
        except AttributeError:
            raise ValueError('Action must be one of {}.'.format(', '.join(Actions.available())))

        self._command_runner = cmd_runner
        self._macros = macros
        self._directives = directives
        self._sequence = sequence
        self._results = []
        self._result = 0
        self._is_setup = False
        self._name = name

    @property
    def sequence(self):
        """The stage execution order."""
        return self._sequence

    @property
    def is_setup(self):
        """True if the setup method was already run, else False."""
        return self._is_setup

    @property
    def name(self):
        """The stage name."""
        return self._name

    def _get_action_function(self, method):
        """Fetches the mapped action function for the provided command runner method.

        :param str method: The command runner method, i.e. provision or teardown.
        :rtype: Callable
        :return: The mapped function.
        """
        method_name = self._action.mapping[method].get(self._command_runner.name, actions.DEFAULT_METHOD)
        method = getattr(actions, method_name)
        return types.MethodType(method, self._command_runner)

    def setup(self):
        """Dynamically set the provision() and teardown() methods for the Command Runner and call it's prepare() method.

        :return: None
        """
        # Dynamically bind the action's provision function to the command runner object.
        self._command_runner.provision = self._get_action_function(actions.SETUP_METHOD)

        # Dynamically bind the action's teardown function to the command runner object.
        self._command_runner.teardown = self._get_action_function(actions.TEARDOWN_METHOD)

        self._is_setup = True

    def run(self, continue_on_fail=False, verbose=False):
        """Executes the commands in the Stage.

        :param bool continue_on_fail: If True, keep running commands even if the last command failed. Default is False.
        :param bool verbose: If True, print the stdout from each Macro status.
        :return: The highest Stage result.
        """
        # Setup if not already setup.
        if not self.is_setup:
            self.setup()

        # Call the provision method.
        try:
            result = self._command_runner.provision()
        except Exception as err:
            raise SetupError(exception=err)
        if not result:
            # TODO: Execute teardown in the case of Vagrant.
            raise SetupError

        # Call the command runner's prepare function.
        self._command_runner.prepare()

        for mac in self._macros:
            directive = self._directives[mac.sequence]

            # Add the prefix to the macro.
            if self._action.add_prefix.get(self._command_runner.name):
                mac.prefix = self._action.add_prefix[self._command_runner.name]

            # Add the suffix to the macro.
            if self._action.add_suffix.get(self._command_runner.name):
                mac.suffix = self._action.add_suffix[self._command_runner.name]

            # Run the macro.
            try:
                _output.log(mode.MACRO_START, directive, mac.command)
                status = self._command_runner.execute(mac)
            except Exception as err:
                raise ExecutionError(exception=err)

            # Handle the result.
            _output.log(mode.MACRO_STATUS, directive, mac.command, status.exit_code)
            self._results.append(status)
            if verbose:
                if status.stdout:
                    _output.print_output(status.stdout)
            if status.exit_code > 0 and not continue_on_fail:
                _output.print_output(status.stderr, is_error=True)
                break

        # Call the teardown method.
        result = self._command_runner.teardown()
        if not result:
            raise TeardownError

        # Set the exit code.
        fails = map(lambda r: True if r.exit_code > 0 else False, self._results)
        if any(fails):
            self._result = 1

        return self._result



"""build-magic

    * Macro - A command to be executed by a CommandRunner.
    * MacroFactory - Generates Macro objects based on the provided commands.
    * CommandRunner - An abstract class for defining methods for executing commands.
"""

__version__ = '0.1.1rc0'

__all__ = [
    'cli',
    'core',
    'macro',
    'output',
    'runner',
    '__version__',
]



"""This module hosts enums and other constants used by build-magic."""
import enum
from enum import Enum, unique
from pathlib import Path
import re

from build_magic.exc import ValidationError


class EnumExt(Enum):
    """Extension for the builtin Enum class."""

    @classmethod
    def names(cls):
        """Provides a tuple of the enum names.

        :rtype: tuple[str]
        :return: A tuple of Enum names.
        """
        return tuple(cls.__members__)

    @classmethod
    def available(cls):
        """Provides the Enum values as a tuple.

        :rtype: tuple[Any]
        :return: A list of available Enum values.
        """
        return tuple([src.value for src in cls])

    @classmethod
    def values(cls):
        """An alias for the available property.

        :rtype: tuple[Any]
        :return: A list of available Enum values.
        """
        return cls.available()


class ParameterMeta(type):
    """Parameter Metaclass for preventing the KEY attribute from being changed."""

    def __new__(mcs, name, bases, attrs):
        """Capture the class attributes and make them immutable."""
        class_ = super().__new__(mcs, name, bases, attrs)
        class_.KEY = attrs.get('KEY', None)
        class_.ENUM = attrs.get('ENUM', None)
        class_.ALIAS = attrs.get('ALIAS', None)
        class_.DEFAULT = attrs.get('DEFAULT', None)
        class_.PATTERN = attrs.get('PATTERN', None)
        return class_


class Parameter(metaclass=ParameterMeta):
    """A class for providing generic key/value pairs."""

    __slots__ = ['_key', '_value']

    def __setattr__(self, key, value):
        """Override the default __setattr__ behavior.

        :param str key: The parameter key to set.
        :param str value: The parameter value to set.
        :return: None
        """
        if key not in ('_key', '_value'):
            raise AttributeError
        super.__setattr__(self, key, value)

    def __init__(self, value=None):
        """Instantiates a new Parameter object.

        :param Any value: The value of the Parameter.
        """
        self._key = self.KEY
        self._value = value or self.default
        if self.pattern:
            if not self.validate_pattern():
                raise ValidationError(message=f'Value {self.value} does not match {self.pattern}.')
        if self.enum:
            if not self.validate_enum():
                valid = (e.value for e in self.enum.__members__.values())
                raise ValidationError(message=f'Value {self.value} is not one of {set(valid)}.')

    def __repr__(self):
        """The string representation of the Parameter object.

        :rtype: str
        :return: The Parameter representation.
        """
        if self.alias:
            return f'<{self.__class__.__name__}: {self._key}, {self.alias}, {self._value}>'
        else:
            return f'<{self.__class__.__name__}: {self._key}, {self._value}>'

    def __eq__(self, other):
        """Compare the object to other to determine if they are equivalent.

        :param Any other: The object to compare against.
        :return: True if the objects are equivalent, otherwise False.
        """
        if isinstance(other, type(self)):
            if self.value == other.value and self.key == other.key:
                return True
        return False

    @property
    def key(self):
        """Getter for the Parameter key.

        :rtype: tuple
        :return: The Parameter object key.
        """
        return self._key

    @property
    def enum(self):
        """Getter for the Parameter enum.

        :rtype: None|enum.Enum
        :return: The Parameter object enum.
        """
        return self.ENUM

    @property
    def alias(self):
        """Getter for the Parameter alias/es.

        :rtype: str|list
        :return: The Parameter object alias.
        """
        return self.ALIAS

    @property
    def default(self):
        """Getter for the Parameter default.

        :rtype: Any
        :return: The Parameter object default.
        """
        return self.DEFAULT

    @property
    def pattern(self):
        """Getter for the Parameter regex pattern.

        :rtype: str
        :return: The Parameter object pattern.
        """
        return self.PATTERN

    @property
    def value(self):
        """Getter for the Parameter value.

        :rtype: Any
        :return: The Parameter object value.
        """
        return self._value

    def as_dict(self):
        """Provides the Parameter object as a dictionary.

        :rtype: dict[Any]
        :return: The Parameter as a dictionary.
        """
        return {self._key: self._value}

    def as_tuple(self):
        """Provides the Parameter object as a tuple where the first value is the key and the second is the value.

        :rtype: tuple[str, Any]
        :return: The Parameter as a tuple.
        """
        return self._key, self._value

    def validate_enum(self):
        """Check to make sure the ENUM attribute is a subclass of enum.enum and value is in ENUM.

        :rtype: bool
        :return: True if the Parameter value is in ENUM, otherwise False.
        """
        if not issubclass(self.enum.__class__, Enum) and type(self.enum) != type(Enum):
            raise TypeError('ENUM is not of type enum.Enum.')
        in_keys = self.value in [e.name for e in self.enum.__members__.values()]
        in_values = self.value in [e.value for e in self.enum.__members__.values()]
        if in_keys:
            self._value = self.enum[self.value].value
        return any((in_keys, in_values))

    def validate_pattern(self):
        """Check to make sure the Parameter value matches the PATTERN attribute.

        :rtype: bool
        :return: True of the Parameter value matches PATTERN, otherwise False.
        """
        if not isinstance(self.pattern, str):
            raise TypeError('PATTERN must be a string.')
        match = re.match(self.pattern, self.value)
        return True if match is not None else False


@unique
class OutputTypes(EnumExt):
    """Mapping of valid output type options to the corresponding Output subclass."""

    BASIC = 'Basic'
    TTY = 'Tty'
    SILENT = 'Silent'


@unique
class Runners(EnumExt):
    """Valid command runner argument names."""

    LOCAL = 'local'
    REMOTE = 'remote'
    VAGRANT = 'vagrant'
    DOCKER = 'docker'


@unique
class Directive(EnumExt):
    """Valid directive argument names."""

    TEST = 'test'
    BUILD = 'build'
    DEPLOY = 'deploy'
    EXECUTE = 'execute'
    INSTALL = 'install'
    RELEASE = 'release'


@unique
class Actions(EnumExt):
    """Valid action argument names."""

    DEFAULT = 'default'
    CLEANUP = 'cleanup'
    PERSIST = 'persist'


@unique
class ExitCode(enum.IntEnum):
    """Valid build-magic exit codes."""

    PASSED = 0
    FAILED = 1
    INTERRUPTED = 4
    INTERNAL_ERROR = 3
    INPUT_ERROR = 2
    NO_TESTS = 5


@unique
class OutputMethod(enum.Enum):
    """Valid build-magic output methods."""

    JOB_START = 'start_job'
    JOB_END = 'end_job'
    STAGE_START = 'start_stage'
    STAGE_END = 'end_stage'
    NO_JOB = 'no_job'
    MACRO_START = 'macro_start'
    MACRO_STATUS = 'macro_status'
    ERROR = 'error'
    INFO = 'info'


class KeyTypes(EnumExt):
    """Valid public/private key types."""

    dsa = 'DSSKey'
    dss = 'DSSKey'
    rsa = 'RSAKey'
    ecdsa = 'ECDSAKey'
    ed25519 = 'Ed25519Key'


class KeyPath(Parameter):
    """Defines a path to a SSH private key or cert file."""

    KEY = 'keypath'
    DEFAULT = str(Path('~/.ssh/id_rsa').expanduser())


class KeyType(Parameter):
    """Defines a SSH public/private key or cert algorithm."""

    KEY = 'keytype'
    DEFAULT = 'RSAKey'
    ENUM = KeyTypes


class KeyPassword(Parameter):
    """Defines the password for decrypting a private key file."""

    KEY = 'key_password'
    ALIAS = 'keypass'


class SSHUser(Parameter):
    """Defines the user for logging into a remote server."""

    KEY = 'ssh_username'
    ALIAS = 'sshuser'


class SSHPassword(Parameter):
    """Defines the password for logging into a remote server."""

    KEY = 'ssh_password'
    ALIAS = 'sshpass'


class HostWorkingDirectory(Parameter):
    """Defines the host machine working directory for Vagrant or Docker runners."""

    KEY = 'hostwd'
    DEFAULT = '.'


class BindDirectory(Parameter):
    """Defines the bind directory of a guest container or VM to the host machine."""

    KEY = 'bind'
    DEFAULT = '/build_magic'



"""This module hosts Actions and functions to dynamically bind to CommandRunner classes."""

import hashlib
import os
import pathlib
import shutil
import subprocess

import docker
from docker.errors import APIError, ImageLoadError
import vagrant


LOCAL = 'local'
REMOTE = 'remote'
VAGRANT = 'vagrant'
DOCKER = 'docker'

SETUP_METHOD = 'provision'
TEARDOWN_METHOD = 'teardown'

BACKUP_PATH = '.build_magic'
TEMP_PATH = '.temp_backup'

DEFAULT_METHOD = 'null'

DARWIN = 'Darwin'
LINUX = 'Linux'
WINDOWS = 'Windows'


def _execute_command(client, cmd):
    """Helper function for executing remote commands.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str cmd: The shell command to execute.
    :rtype: tuple
    :return: A tuple of paramiko.channel.ChannelFile object representing stdin, stdout, and stderr.
    """
    return client.exec_command(cmd)


def _get_remote_os(client):
    """Attempt to get the operating system of a remote machine.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :rtype: str|None
    :return: The operating system of the remote machine or None if detection fails.
    """
    system = None
    stdin, stdout, stderr = _execute_command(client, 'uname')
    if stdout.channel.recv_exit_status() == 0:
        os_ = stdout.readlines()[0].strip()
        if os_ in (LINUX, DARWIN):
            system = os_
        elif os_.startswith(WINDOWS):
            system = WINDOWS
    else:
        # Check if we're connecting to Windows.
        stdin, stdout, stderr = _execute_command(client, '%OS%')
        os_ = stdout.readlines()[0].strip()
        if os_ == 'Windows_NT':
            system = WINDOWS
    return system


def _get_files_and_hashes(client, directory=''):
    """Helper function for getting file names and their corresponding SHA1 hashes from a unix-like file system.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str directory: The directory to recursively fetch file names and hashes from.
    :rtype: paramiko.channel.ChannelFile
    :return: The resulting stdout object from the executed command.
    """
    path = directory or '$PWD'
    cmd = f'find {path} -type f | xargs shasum {path}/*'
    stdin, stdout, stderr = _execute_command(client, cmd)
    return stdout


def _get_files_unix(client, working_directory=''):
    """Helper function for getting a list of files on a remote unix-like file system.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str working_directory: The directory to recursively fetch file names from.
    :rtype: tuple[paramiko.channel, list[str]]
    :return: A tuple of the SSH channel and each line of the stdout object from the executed command as a list.
    """
    path = working_directory or '$PWD'
    cmd = f'find {path} -type f'
    stdin, stdout, stderr = _execute_command(client, cmd)
    return stdout.channel, [f for f in stdout.readlines()]


def _get_files_windows(client, working_directory=''):
    """Helper function for getting a list of files on a remote Windows system.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str working_directory: The directory to recursively fetch file names from.
    :rtype: tuple[paramiko.channel, list[str]]
    :return: A tuple of the SSH channel and each line of the stdout object from the executed command as a list.
    """
    path = working_directory + ' ' if working_directory else working_directory
    cmd = f'dir {path}/a-D /S /B'
    stdin, stdout, stderr = _execute_command(client, cmd)
    return stdout.channel, [f for f in stdout.readlines()]


def _get_directories_unix(client, working_directory=''):
    """Helper function for getting a list of directories on a remote unix-like file system.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str working_directory: The directory to recursively fetch directory names from.
    :rtype: tuple[paramiko.channel, list[str]]
    :return: A tuple of the SSH channel and each line of the stdout object from the executed command as a list.
    """
    path = working_directory or '$PWD'
    cmd = f'find {path} -type d'
    stdin, stdout, stderr = _execute_command(client, cmd)
    return stdout.channel, [d for d in stdout.readlines() if d != path]


def _get_directories_windows(client, working_directory=''):
    """Helper function for getting a list of directories on a remote Windows file system.

    :param paramiko.SSHClient client: The SSHClient object to use for executing the command.
    :param str working_directory: The directory to recursively fetch directory names from.
    :rtype: tuple[paramiko.channel, list[str]]
    :return: A tuple of the SSH channel and each line of the stdout object from the executed command as a list.
    """
    path = working_directory + ' ' if working_directory else working_directory
    cmd = f'dir {path}/AD /B /ON /S'
    stdin, stdout, stderr = _execute_command(client, cmd)
    return stdout.channel, [d for d in stdout.readlines()]


def _parse_files(file_list):
    """Helper function to parse a list of files into a tuple for storage.

    :param list[str] file_list: A list of file names to parse and format.
    :return: A list of tuples where the first value is the file name and the second value is None.
    """
    return [(file.strip(), None) for file in file_list if file]


def _parse_directories(dir_list):
    """Helper function for parsing and cleaning a list of directories for analysis.

    :param list[str] dir_list: A list of directories to parse.
    :return: The cleaned up list of directories.
    """
    return [line for line in dir_list if line is not None]


def _get_local_files_and_directories(files):
    """Helper function for getting hashes of local files, as well as subdirectories.

    :param files: Every "file" under a directory.
    :rtype: tuple[tuple[str, str], str]
    :return: A tuple of the file hashes and subdirectories.
    """
    file_hashes, dirs = [], []
    for file in sorted(files):
        try:
            file_hashes.append((str(file), hashlib.sha1(pathlib.Path(file).read_bytes()).hexdigest()))
        except IsADirectoryError:
            dirs.append(str(file))
    return file_hashes, dirs


# def _clear_directory(directory):
#     """
#
#     :param pathlib.Path directory:
#     :return:
#     """
#     for file in directory.iterdir():
#         if file.is_dir():
#             if len(list(file.iterdir())) > 0:
#                 _clear_directory(directory)
#             else:
#                 file.rmdir()
#         else:
#             file.unlink()


class Action:
    """An Action is used to dynamically define the setup and teardown process for a CommandRunner.
    At a minimum, a Default action is called to specify and dynamically bind the setup and teardown methods.
    New functions can be added to Action classes to modify the setup and teardown behavior of a CommandRunner.
    Additionally, commands can be added as macro prefixes and suffixes for each CommandRunner.

    The mapping class attribute is a dictionary with keys for the setup and teardown methods.
    Each key then has a dictionary of key/value pairs, where the key the CommandRunner name,
    and the value is the function name to dynamically bind to the setup or teardown method. For example,

        mapping = {
            'provision': {
                'local': 'copy_from',
                'remote': 'copy_to',
            }
        }

    will map the function "copy_from()" to the "provision()" method of the Local CommandRunner class
    and the function "copy_from()" to the "provision()" method of the Remote CommandRunner class.

    The add_prefix and add_suffix class attributes are dictionaries where the key is the CommandRunner name
    and the value is the function name to dynamically add as a prefix or suffix to each macro being executed
    by the corresponding CommandRunner class. For example,

        add_prefix = {
            'vagrant': 'cd /',
        }

    will add the command "cd /" as a prefix to each command executed by the Vagrant CommandRunner class.
    """

    # Class attribute that maps a command runner to a command that will be added before each executed command.
    add_prefix = {}

    # Class attribute that maps a command runner to a command that will be added after each executed command.
    add_suffix = {}

    # Class attribute that defines a mapping between command runners and the functions to run for setup and teardown.
    mapping = {
        SETUP_METHOD: {},
        TEARDOWN_METHOD: {},
    }


class Default(Action):
    """The default action which executes commands in an environment individually."""

    mapping = {
        SETUP_METHOD: {
            LOCAL: DEFAULT_METHOD,
            REMOTE: DEFAULT_METHOD,
            DOCKER: 'container_up',
            VAGRANT: 'vm_up',
        },
        TEARDOWN_METHOD: {
            LOCAL: DEFAULT_METHOD,
            REMOTE: DEFAULT_METHOD,
            DOCKER: 'container_destroy',
            VAGRANT: 'vm_destroy',
        },
    }

    add_prefix = {
        VAGRANT: 'cd /vagrant;',
    }


class Cleanup(Action):
    """Action for deleting any files created while executing a Stage."""

    mapping = {
        SETUP_METHOD: {
            LOCAL: 'capture_dir',
            REMOTE: 'remote_capture_dir',
            DOCKER: 'docker_capture_dir',
            VAGRANT: 'vm_up',
        },
        TEARDOWN_METHOD: {
            LOCAL: 'delete_new_files',
            REMOTE: 'remote_delete_files',
            DOCKER: 'docker_delete_new_files',
            VAGRANT: 'vm_destroy',
        },
    }

    add_prefix = {
        VAGRANT: 'cd /vagrant;',
    }


# class Restore(Action):
#     """Action for restoring the working directory after executing a Stage."""
#
#     mapping = {
#         SETUP_METHOD: {
#             LOCAL: 'backup_dir',
#             DOCKER: 'backup_dir',
#             VAGRANT: 'vm_up',
#         },
#         TEARDOWN_METHOD: {
#             LOCAL: 'restore_from_backup',
#             DOCKER: 'restore_from_backup',
#             VAGRANT: 'vm_destroy',
#         },
#     }
#
#     add_prefix = {
#         VAGRANT: 'cd /vagrant;',
#     }


class Persist(Action):
    """Action for starting an environment but doesn't teardown."""

    mapping = {
        SETUP_METHOD: {
            LOCAL: DEFAULT_METHOD,
            REMOTE: DEFAULT_METHOD,
            DOCKER: 'container_up',
            VAGRANT: 'vm_up',
        },
        TEARDOWN_METHOD: {
            LOCAL: DEFAULT_METHOD,
            REMOTE: DEFAULT_METHOD,
            DOCKER: DEFAULT_METHOD,
            VAGRANT: DEFAULT_METHOD,
        },
    }

    add_prefix = {
        VAGRANT: 'cd /vagrant;',
    }


def null(self):
    """Basic No-op."""
    return True


def vm_up(self):
    """Starts up a VM according to the corresponding Vagrantfile."""
    self._vm = vagrant.Vagrant()
    try:
        self._vm.up()
    except subprocess.CalledProcessError as err:
        print(str(err))
        self.teardown()
        return False
    return True


def vm_destroy(self):
    """Destroys the VM used for executing commands."""
    if hasattr(self, '_vm') and isinstance(self._vm, vagrant.Vagrant):
        try:
            self._vm.destroy()
        except subprocess.CalledProcessError as err:
            print(str(err))
            return False
    return True


def capture_dir(self):
    """Capture a list of all the files and their MD5 hashes in a directory."""
    try:
        pwd = pathlib.Path(self.working_directory).resolve()
        matches = pwd.rglob('*')
        self._existing_files, self._existing_dirs = _get_local_files_and_directories(matches)
    except:
        return False
    return True


def delete_new_files(self):
    """Deletes all files not previously captured."""
    result = False
    if hasattr(self, '_existing_files') and isinstance(self._existing_files, list):
        if len(self._existing_files) > 0:
            files, hashes = zip(*self._existing_files)
            pwd = pathlib.Path(self.working_directory).resolve()
            current = []
            for file in sorted(pwd.rglob('*')):
                try:
                    current.append((file, hashlib.sha1(pathlib.Path(file).read_bytes()).hexdigest()))
                except IsADirectoryError:
                    continue
            _, new_hashes = zip(*current)
            for file, hash_ in current:
                # Remove any new files.
                if str(file) not in files and hash_ not in hashes:
                    os.remove(file)
                    continue
                # Remove any files copied from existing files.
                elif hash_ in hashes and (str(file), hash_) not in self._existing_files and new_hashes.count(hash_) > 1:
                    os.remove(file)
                    continue
            result = True
    if hasattr(self, '_existing_dirs') and isinstance(self._existing_dirs, list):
        pwd = pathlib.Path(self.working_directory).resolve()
        for file in sorted(pwd.rglob('*'), reverse=True):
            if file.is_dir() and file not in self._existing_dirs:
                try:
                    os.rmdir(file)
                except OSError:
                    continue
        result = True
    return result


def docker_capture_dir(self):
    """Capture a list of all the files and their MD5 hashes in the host directory."""
    try:
        pwd = pathlib.Path(self.host_wd).resolve()
        matches = pwd.rglob('*')
        self._existing_files, self._existing_dirs = _get_local_files_and_directories(matches)
    except:
        return False
    return container_up(self)


def docker_delete_new_files(self):
    """Deletes all files not previously captured in the host directory."""
    result = False
    if hasattr(self, '_existing_files') and isinstance(self._existing_files, list):
        if len(self._existing_files) > 0:
            files, hashes = zip(*self._existing_files)
            pwd = pathlib.Path(self.host_wd).resolve()
            current = []
            for file in sorted(pwd.rglob('*')):
                try:
                    current.append((file, hashlib.sha1(pathlib.Path(file).read_bytes()).hexdigest()))
                except IsADirectoryError:
                    continue
            _, new_hashes = zip(*current)
            for file, hash_ in current:
                # Remove any new files.
                if str(file) not in files and hash_ not in hashes:
                    os.remove(file)
                    continue
                # Remove any files copied from existing files.
                elif hash_ in hashes and (str(file), hash_) not in self._existing_files and new_hashes.count(hash_) > 1:
                    os.remove(file)
                    continue
            result = True
    if hasattr(self, '_existing_dirs') and isinstance(self._existing_dirs, list):
        pwd = pathlib.Path(self.host_wd).resolve()
        for file in sorted(pwd.rglob('*'), reverse=True):
            if file.is_dir() and file not in self._existing_dirs:
                try:
                    os.rmdir(file)
                except OSError:
                    continue
        result = True
    container_res = container_destroy(self)
    return all([result, container_res])


def backup_dir(self):
    """Creates a backup of the working directory."""
    wd = pathlib.Path(self.working_directory).resolve()
    backup = wd / BACKUP_PATH
    try:
        if backup.exists():
            shutil.rmtree(backup)
        shutil.copytree(wd, backup)
    except (PermissionError, Exception):
        return False
    return True


def restore_from_backup(self):
    """Restore from the backup directory and delete it."""
    wd = pathlib.Path(self.working_directory).resolve()
    backup = wd / BACKUP_PATH
    if backup.exists():
        try:
            temp = wd.parent / TEMP_PATH
            shutil.move(wd, temp)
            shutil.copytree(temp / BACKUP_PATH, wd)
            shutil.rmtree(temp)
            return True
        except (PermissionError, Exception):
            return False
    else:
        return False


# def remote_backup_dir(self):
#     """Create a backup directory on a remote file system."""
#     wd = pathlib.Path(self.working_directory).resolve()
#     backup = wd / BACKUP_PATH
#     client = self.connect()
#     cmd = 'uname'
#     # Try to get the OS of the remote system.
#     stdin, stdout, stderr = client.exec_command(cmd)
#     if stdout.channel.recv_exit_status() == 0:
#         stdin, stdout, stderr = client.exec_command(f'cd {str(wd)}')
#         if stdout.readlines()[0] in ('Linux', 'Darwin'):
#             stdin, stdout, stderr = client.exec_command(f'ls {str(backup)}')
#             if stdout.channel.recv_exit_status() != 0:
#                 stdin, stdout, stderr = client.exec_command(f'rm -R -f {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'mkdir {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'cp -R * {str(backup)}')
#         elif stdout.readlines()[0].startswith('Windows'):
#             stdin, stdout, stderr = client.exec_command(f'ls {str(backup)}')
#             if stdout.channel.recv_exit_status() != 0:
#                 stdin, stdout, stderr = client.exec_command(f'del /q {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'mkdir {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'xcopy /y * {str(backup)}')
#         else:
#             return False
#     else:
#         # Check if we're connecting to Windows.
#         cmd = '%OS%'
#         stdin, stdout, stderr = client.exec_command(cmd)
#         if stdout.readlines()[0] == 'Windows_NT':
#             stdin, stdout, stderr = client.exec_command(f'cd {str(wd)}')
#             stdin, stdout, stderr = client.exec_command(f'ls {str(backup)}')
#             if stdout.channel.recv_exit_status() != 0:
#                 stdin, stdout, stderr = client.exec_command(f'del /q {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'mkdir {str(backup)}')
#             stdin, stdout, stderr = client.exec_command(f'xcopy /y * {str(backup)}')
#         else:
#             return False
#     return True


def remote_restore_from_backup(self):
    """Restore from a backup directory on a remote file system."""


def remote_capture_dir(self):
    """Capture a list of all the files in a directory on a remote system."""
    client = self.connect()

    # Try to get the OS of the remote system.
    system = _get_remote_os(client)
    if not system:
        # OS type isn't supported yet.
        return False

    if system != WINDOWS:
        # Try to get the filename and SHA1 hashes of the remote working directory.
        stdout = _get_files_and_hashes(client, self.working_directory)

        if stdout.channel.recv_exit_status() == 0:
            existing = []
            for line in stdout.readlines():
                if line:
                    hash_, file = tuple(line.split('  '))
                    existing.append((file, hash_))
                else:
                    continue
            self._existing_files = existing
        # If trying to get the hashes fails, just use the filenames.
        else:
            _, files = _get_files_unix(client, self.working_directory)
            self._existing_files = _parse_files(files)

        # Try to get the sub-directories of the remote working directory.
        _, dirs = _get_directories_unix(client, self.working_directory)
        self._existing_dirs = _parse_directories(dirs)
    elif system == WINDOWS:
        channel, files = _get_files_windows(client, self.working_directory)

        if channel.recv_exit_status() == 0:
            self._existing_files = _parse_files(files)
        else:
            return False
        _, dirs = _get_directories_windows(client, self.working_directory)
        self._existing_dirs = _parse_directories(dirs)
    return True


def remote_delete_files(self):
    """Deletes all files not previously captured on a remote system."""
    current_files = []
    current_dirs = []
    client = self.connect()
    # Try to get the OS of the remote system.
    system = _get_remote_os(client)
    if not system:
        # OS type isn't supported yet.
        return False

    if system != WINDOWS:
        # Try to get the filename and SHA1 hashes of the remote working directory.
        stdout = _get_files_and_hashes(client, self.working_directory)
        if stdout.channel.recv_exit_status() == 0:
            for line in stdout.readlines():
                if line:
                    hash_, file = tuple(line.split('  '))
                    current_files.append((file, hash_))
                else:
                    continue
        # If trying to get the hashes fails, just use the filenames.
        else:
            channel, files = _get_files_unix(client, self.working_directory)
            if channel.recv_exit_status() == 0:
                current_files = _parse_files(files)
            else:
                # Cannot get the filenames.
                return False
        # Try to get the sub-directories of the remote working directory.
        _, dirs = _get_directories_unix(client, self.working_directory)
        current_dirs = _parse_directories(dirs)
    elif system == WINDOWS:
        channel, files = _get_files_windows(client, self.working_directory)
        if channel.recv_exit_status() == 0:
            current_files = _parse_files(files)
        else:
            # Cannot get the filenames.
            return False
        _, dirs = _get_directories_windows(client, self.working_directory)
        current_dirs = _parse_directories(dirs)

    to_delete = []
    if current_files:
        _, new_hashes = zip(*set(current_files))
        if hasattr(self, '_existing_files') and len(self._existing_files) > 0:
            files, hashes = zip(*self._existing_files)
        else:
            files, hashes = [], []
        for file, hash_ in current_files:
            if str(file) not in files and hash_ not in hashes:
                if self.working_directory:
                    file = pathlib.Path(self.working_directory) / file
                to_delete.append(str(file).strip('\n'))
                continue
            elif hash_ in hashes and (str(file), hash_) not in self._existing_files and new_hashes.count(hash_) > 1:
                if self.working_directory:
                    file = pathlib.Path(self.working_directory) / file
                to_delete.append(str(file).strip('\n'))
                continue
        if to_delete:
            _execute_command(client, f'rm {" ".join(to_delete)}')
    if system != WINDOWS:
        to_delete = []
        if current_dirs:
            if hasattr(self, '_existing_dirs') and len(self._existing_dirs) > 0:
                for directory in sorted(current_dirs, reverse=True):
                    if directory not in self._existing_dirs:
                        to_delete.append(directory.strip('\n'))
            if to_delete:
                _execute_command(client, f'rm -rf {" ".join(to_delete)}')
    result = True
    return result


def container_up(self):
    """Starts up a new container based on the image set in self.environment."""
    self.client = docker.from_env()
    try:
        self.container = self.client.containers.run(
            self.environment,
            detach=True,
            tty=True,
            entrypoint='sh',
            working_dir=self.working_directory,
            mounts=[self.binding],
            name='build-magic',
        )
    except (APIError, AttributeError, ImageLoadError):
        return False
    return True


def container_destroy(self):
    """Destroys the container used for executing commands."""
    try:
        self.container.kill()
        self.container.remove()
        return True
    except (APIError, Exception):
        return False



from setuptools import setup

from build_magic import __version__ as version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='build-magic',
    version=version,
    packages=[
        'build_magic',
    ],
    url='https://cmmorrow.github.io/build-magic/',
    license='MIT',
    author='Chris Morrow',
    author_email='cmmorrow@gmail.com',
    description='A general purpose build/install/deploy tool.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='!=2.*, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, >=3.6',
    project_urls={
        "Documentation": 'https://cmmorrow.github.io/build-magic/',
        "Source": 'https://github.com/cmmorrow/build-magic',
        "Bug Tracker": 'https://github.com/cmmorrow/build-magic/issues',
    },
    entry_points={
        'console_scripts': [
            'build-magic=build_magic.cli:build_magic',
        ],
    },
    package_data={
        'build_magic': [
            'static/config_schema.json',
        ],
    },
    install_requires=[
        'click==7.1.2',
        'colorama==0.4.4',
        'jsonschema==3.2.0',
        'docker==4.3.1',
        'paramiko==2.7.2',
        'python-vagrant==0.5.15',
        'PyYAML==5.4',
        'scp==0.13.3',
    ],
    tests_require=[
        'pytest',
        'freezegun',
    ],
    classifers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
    ],
)



"""Module for defining CommandRunner classes."""

import os
from pathlib import Path
import re
import shutil
import socket
import subprocess

from docker.types import Mount
from docker.errors import ContainerError
import paramiko
from scp import SCPClient

from build_magic.reference import BindDirectory, HostWorkingDirectory


class Status:
    """The captured output of a command executed by a CommandRunner."""

    __slots__ = ['_stdout', '_stderr', '_exit_code']

    def __init__(self, stdout='', stderr='', exit_code=0):
        """Instantiates a new Status object.

        :param bytes|str stdout: The stdout captured by a CommandRunner.
        :param bytes|str stderr: The stderr captured by a CommandRunner.
        :param num exit_code: The exit code captured by a CommandRunner.
        """
        self._stdout = stdout
        self._stderr = stderr
        self._exit_code = exit_code

    @property
    def stdout(self):
        """Provides the captured stdout.

        :rtype: str
        :return: The captured stdout.
        """
        return self._stdout

    @property
    def stderr(self):
        """Provides the captured stderr.

        :rtype: str
        :return: The captured stderr.
        """
        return self._stderr

    @property
    def exit_code(self):
        """The captured exit code.

        :rtype: int
        :return: The captured exit code.
        """
        return self._exit_code

    @staticmethod
    def _validate(other):
        """Raise an exception if other isn't a Status object.

        :param Any other: The object to validate.
        :return: None
        """
        if not isinstance(other, Status):
            raise TypeError(f'Cannot compare object of type Status with object of type {type(other)}')

    def __repr__(self):
        """Overrides the default string representation to display stdout, stderr, and exit_code.

        :rtype: str
        :return: The Status string representation.
        """
        return f'<stdout={self.stdout}, stderr={self.stderr}, exit_code={self.exit_code}>'

    def __eq__(self, other):
        """Compare two Status object to see if they are equal.

        :param Any other: The Status object to compare against.
        :rtype: bool
        :return: True if the Status objects are equal, otherwise False.
        """
        self._validate(other)
        return self.stdout == other.stdout and self.stderr == other.stderr and self.exit_code == other.exit_code

    def __lt__(self, other):
        """Compare if the exit code of the current Status object is less than the exit code of another Status object.

        :param Any other: The Status object to compare against.
        :rtype: bool
        :return: True if the exit code is less than the exit code of other, else False.
        """
        self._validate(other)
        return self.exit_code < other.exit_code

    def __le__(self, other):
        """Compare if the exit code of the current Status object is less than or equal to the exit code of another
        Status object.

        :param Any other: The Status object to compare against.
        :rtype: bool
        :return: True if the exit code is less than or equal to the exit code of other, else False.
        """
        self._validate(other)
        return self.exit_code <= other.exit_code

    def __gt__(self, other):
        """Compare if the exit code of the current Status object is greater than the exit code of another Status object.

        :param Any other: The Status object to compare against.
        :rtype: bool
        :return: True if the exit code is greater than the exit code of other, else False.
        """
        self._validate(other)
        return self.exit_code > other.exit_code

    def __ge__(self, other):
        """Check if the exit code of the Status object is greater than or equal to the exit code of another Status
        object.

        :param Any other: The Status object to compare against.
        :rtype: bool
        :return: True if the exit code is greater than or equal to the exit code of other, else False.
        """
        self._validate(other)
        return self.exit_code >= other.exit_code


class CommandRunner:
    """An abstract class for defining methods for executing commands.

    When instantiating and calling a CommandRunner child object,
    the object's methods should be called in the following order:

        - prepare()
        - provision()
        - execute()
        - teardown()
    """

    def __init__(self, environment, working_dir='', copy_dir='', timeout=30, artifacts=None, parameters=None):
        """Instantiate a new CommandRunner object."""
        self.environment = environment
        self.working_directory = working_dir
        self.copy_from_directory = copy_dir
        self.timeout = timeout
        self.artifacts = [] if not artifacts else artifacts
        self.parameters = {} if not parameters else parameters

        # TODO: Need to validate the types for artifacts and parameters.

    @staticmethod
    def cd(directory):
        """Changes the current working directory.

        :param str|Path directory: The directory to change to.
        :rtype: bool
        :return: True if directory is not an empty string or None.
        """
        if directory:
            path = Path(directory)
            try:
                os.chdir(path)
            except (OSError, Exception):
                return False
            return True
        else:
            return False

    def copy(self, src, dst):
        """Copies the CommandRunner object's artifacts from one path to another.

        :param str|Path src: The source path to copy from.
        :param str|Path dst: The destination path to copy to.
        :rtype: bool
        :return: True if src and dst are not empty strings or None.
        """
        if not self.artifacts:
            return False
        if src and dst:
            for artifact in self.artifacts:
                if src == '.':
                    src = Path.cwd().resolve()
                source = Path(src) / artifact
                try:
                    shutil.copy(source, dst)
                except (OSError, Exception):
                    return False
            return True
        else:
            return False

    def provision(self):
        """Base provision() method.

        This method should be overridden by dynamically assigning it to a child object.

        :rtype: Any
        :return: The result of the callable.
        """
        return False

    def execute(self, macro):
        """Base execute() method.

        :param Macro macro: The Macro object to execute.
        :return: None
        """
        raise NotImplementedError

    def prepare(self):
        """Base prepare() method.

        :return: None
        """
        raise NotImplementedError

    def teardown(self):
        """Base provision() method.

        This method should be overridden by dynamically assigning it to a child object.

        :rtype: Any
        :return: The result of the callable.
        """
        return False

    @property
    def name(self):
        """Provides the CommandRunner name."""
        return str(self.__class__.__name__).lower()


class Local(CommandRunner):
    """Manages macros executed on the local host machine."""

    def __init__(self, environment='', working_dir='', copy_dir='', timeout=30, artifacts=None, parameters=None):
        """Instantiates a new Local command runner object."""
        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)

    def prepare(self):
        """Changes to the specified working directory and copies artifacts if necessary.

        :rtype: bool
        :return: Returns True.
        """
        if self.copy_from_directory:
            self.copy(self.copy_from_directory, self.working_directory)
        if self.working_directory:
            self.cd(self.working_directory)
        return True

    def execute(self, macro):
        """Executes the Macro object on the local host machine.

        :param Macro macro: The Macro object the execute.
        :rtype: Status
        :return: A Status object reflecting the results of the macro.
        """
        command = macro.as_string()
        result = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            # timeout=self.timeout,
            shell=True,
        )
        return Status(result.stdout, result.stderr, result.returncode)


class Remote(CommandRunner):
    """Manages macros executed on a remote host machine."""

    _default_ssh_path = Path('~/.ssh').expanduser()
    _environment_pattern = r'^(?:([\w\-.]+)@)?([\w\-.]+)(?::([0-9]{2,5}))?$'

    def __init__(
            self,
            environment='localhost',
            working_dir='',
            copy_dir='',
            timeout=30,
            artifacts=None,
            parameters=None,
    ):
        """Instantiates a new Remote command runner object."""
        if working_dir == '.':
            working_dir = ''

        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)

        self.user = None
        self.host = None
        self.port = None
        match = re.match(self._environment_pattern, environment)
        if match:
            self.user, self.host = match.group(1), match.group(2)
            self.port = int(match.group(3)) if match.group(3) else 22

        self.key = self._get_ssh_key()

    def _get_ssh_key(self):
        """Get the SSH private key to use for authentication.

        :rtype: paramiko.PKey
        :return: The SSH private key object.
        """
        # Get the SSH key type from parameters or default to RSA.
        key_type_param = self.parameters.get('keytype')
        if key_type_param:
            key_type = getattr(paramiko, f'{key_type_param.value}')
        else:
            key_type = paramiko.RSAKey

        # Get the SSH private key path from parameters or default to $HOME/.ssh/id_rsa.
        key_path_param = self.parameters.get('keypath')
        if key_path_param:
            key_path = key_path_param.value
        else:
            key_path = str(self._default_ssh_path / 'id_rsa')

        # Get the SSH private key password from parameters if provided.
        if 'keypass' in self.parameters:
            key_pass_param = self.parameters.get('keypass').value
        elif 'key_password' in self.parameters:
            key_pass_param = self.parameters.get('key_password').value
        else:
            key_pass_param = None
        try:
            return key_type.from_private_key_file(key_path, password=key_pass_param)
        except (
            FileNotFoundError,
            IOError,
            Exception,
            paramiko.ssh_exception.SSHException,
            paramiko.PasswordRequiredException,
        ) as err:
            raise ValueError(f'SSH failure: {err}')

    def connect(self):
        """Creates an SSH connection.

        :rtype: paramiko.SSHClient
        :return: The SSH client.
        """
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        conn_args = dict(hostname=self.host, pkey=self.key)
        if self.port:
            conn_args.update(dict(port=self.port))
        if self.user:
            conn_args.update(dict(username=self.user))
        try:
            ssh.connect(**conn_args)
        except socket.gaierror:
            raise Exception('SSH connection failed.')
        return ssh

    def copy(self, src, dst=None):
        """Copies the object's artifacts from a directory on the local host to a directory on the remote host.

        :param str|Path src: The source directory to copy from.
        :param str|Path dst: The destination directory to copy to.
        :rtype: bool
        :return: True if the copy operation was successful.
        """
        # Add the absolute path to each artifact.
        src = Path(src)
        files = []
        for artifact in self.artifacts:
            files.append(str(src / artifact))

        if not files:
            return False

        # Connect to the remote host over scp.
        ssh = self.connect()

        # Copy each file to the remote host.
        with SCPClient(ssh.get_transport()) as cp:
            if dst:
                cp.put(files, remote_path=dst)
            else:
                cp.put(files)
        return True

    def prepare(self):
        """Handles copying artifacts to the remote host if necessary.

        :rtype: bool
        :return: True if copying artifacts was initiated and successful, otherwise False.
        """
        if self.copy_from_directory:
            src = self.copy_from_directory
            if self.working_directory:
                dst = self.working_directory
            else:
                dst = None
            return self.copy(src, dst)
        else:
            return False

    def execute(self, macro):
        """Executes the Macro object on the remote host machine.

        :param Macro macro: The Macro object to execute.
        :rtype: Status
        :return: A Status object that reflects the results of the macro.
        """
        if self.working_directory:
            macro.prefix = f'cd {self.working_directory};'
        command = macro.as_string()

        ssh = self.connect()
        try:
            stdin_, stdout_, stderr_ = ssh.exec_command(command, timeout=self.timeout)
            stdout = stdout_.readlines()
            stderr = stderr_.readlines()
            exit_code = stdout_.channel.recv_exit_status()
        except socket.timeout:
            raise TimeoutError(
                'Connection to remote host {} timed out after {} seconds.'.format(self.host, self.timeout)
            )
        finally:
            ssh.close()
        return Status(stdout=''.join(stdout), stderr=''.join(stderr), exit_code=exit_code)


class Vagrant(CommandRunner):
    """Manages macros executed executed in a guest virtual machine managed by Vagrant."""

    def __init__(
            self,
            environment='.',
            working_dir='/home/vagrant',
            copy_dir='',
            timeout=30,
            artifacts=None,
            parameters=None,
    ):
        """Instantiates a new Vagrant command runner object."""
        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
        self._vm = None
        if self.environment == 'Vagrantfile':
            self.environment = '.'
        if self.environment != '.':
            if 'Vagrantfile' in self.environment:
                self.environment = self.environment.split('Vagrantfile')[0]
            os.environ.pop('VAGRANT_CWD', '')
            os.environ['VAGRANT_CWD'] = str(Path(self.environment).resolve())
        self.host_wd = self.parameters.get('hostwd', HostWorkingDirectory(self.environment)).value
        self.bind_path = self.parameters.get('bind', BindDirectory('/vagrant')).value

    def prepare(self):
        """Handles copying artifacts to the working directory if necessary."""
        if self._vm and self.working_directory != self.bind_path:
            try:
                self._vm.ssh(command=f'sudo mkdir {self.working_directory}')
            except subprocess.CalledProcessError:
                return False
        if self.copy_from_directory and len(self.artifacts) > 0:
            self.copy(Path(self.copy_from_directory).resolve(), Path(self.host_wd).resolve())
            try:
                self._vm.ssh(command=f'sudo cp -R {self.bind_path}/* {self.working_directory}')
            except subprocess.CalledProcessError:
                return False
            return True
        return False

    def execute(self, macro):
        """Executes the Macro in the VM.

        :param Macro macro: The Macro object to execute.
        :return: The Status of the executed Macro.
        """
        if self.working_directory != '/home/vagrant':
            macro.prefix = f'cd {self.working_directory};'
        cmd = macro.as_string()
        try:
            out = self._vm.ssh(command=cmd)
        except subprocess.CalledProcessError as err:
            return Status('', str(err), 1)
        return Status(out, '', 0)


class Docker(CommandRunner):
    """Manages macros executed in a Docker container."""

    def __init__(
            self,
            environment='alpine',
            working_dir='/build_magic',
            copy_dir='',
            timeout=30,
            artifacts=None,
            parameters=None
    ):
        """Instantiates a new Docker command runner object."""
        super().__init__(environment, working_dir, copy_dir, timeout, artifacts, parameters)
        if self.working_directory == '.':
            self.working_directory = '/build_magic'
        self.host_wd = self.parameters.get('hostwd', HostWorkingDirectory('.')).value
        self.bind_path = self.parameters.get('bind', BindDirectory()).value
        self.binding = Mount(
            target=self.bind_path,
            source=str(Path(self.host_wd).resolve()),
            type='bind',
        )
        self.client = None
        self.container = None

    def prepare(self):
        """Handles copying artifacts to the working directory if necessary."""
        if self.copy_from_directory:
            if self.working_directory != self.bind_path and len(self.artifacts) > 0:
                self.copy(self.copy_from_directory, Path(self.host_wd).resolve())
                if not self.container:
                    self.provision()
                try:
                    self.container.exec_run(cmd=f'mkdir {self.working_directory}')
                    self.container.exec_run(cmd=f'/bin/sh -c "cp -R {self.bind_path}/* {self.working_directory}"')
                except ContainerError:
                    return False
                return True
        return False

    def execute(self, macro):
        """Executes the Macro in the container.

        :param Macro macro: The Macro object to execute.
        :return: The Status of the executed Macro.
        """
        if macro:
            command = ['/bin/sh', '-c', macro.as_string()]
        else:
            command = None

        if not self.container:
            self.provision()
        try:
            code, (out, err) = self.container.exec_run(
                cmd=command,
                stdout=True,
                stderr=True,
                tty=True,
                demux=True,
            )
            status = Status(stdout=out, stderr=err, exit_code=code)
        except ContainerError as err:
            status = Status('', stderr=str(err), exit_code=1)
        return status



"""This module hosts the build-magic exception."""


class BuildMagicException(Exception):
    """build-magic base exception class."""

    msg = 'build-magic error'

    def __init__(self, exception=None, message=''):
        """Instantiates a BuildMagicException object.

        :param Exception exception: An exception to wrap.
        :param str message: An error message to display.
        """
        if exception:
            super().__init__(f'{self.msg}: {str(exception)}')
        elif message:
            super().__init__(f'{self.msg}: {message}')
        else:
            super().__init__(f'{self.msg}.')


class ExecutionError(BuildMagicException):
    """A general build-magic execution error."""

    msg = 'Command execution error'


class SetupError(BuildMagicException):
    """An error when setting up a CommandRunner."""

    msg = 'Setup failed'


class TeardownError(BuildMagicException):
    """An error when tearing down a CommandRunner."""

    msg = 'Teardown failed'


class NoJobs(BuildMagicException):
    """There are no jobs to execute."""

    msg = 'No jobs to execute'


class ValidationError(BuildMagicException):
    """Parameter validation failed."""

    msg = 'Validation failed'
